"""
Data support functionality like retrieving USGS gage data and parsing stage files.

Copyright (C) 2020 Daniel Philippus.
Full copyright notice in main.py.
"""

from urllib.request import urlopen


def usgsURL(gage, end=None, start=None, period=None):
    """
    Generate USGS gage data url.
    :param gage: gage number
    :param start: start date (yyyy-mm-dd)
    :param end: end date
    :param period: number of days to retrieve (int)
    :return: the URL
    """
    # Period: https://nwis.waterdata.usgs.gov/ca/nwis/uv/?cb_00060=on&cb_00065=on&format=rdb&site_no=09429100&period=500&begin_date=2020-02-17&end_date=2020-02-24
    # Date range: https://waterdata.usgs.gov/ca/nwis/uv?cb_00060=on&cb_00065=on&format=rdb&site_no=09423350&period=&begin_date=2020-02-17&end_date=2020-02-24
    # format order: site_no, period, begin_date, end_date - all strings
    # Period, begin_date can each be left empty if the other is specified
    base = "https://nwis.waterdata.usgs.gov/ca/nwis/uv/?cb_00060=on&cb_00065=on&format=rdb&site_no=%s&period=%s&begin_date=%s&end_date=%s"
    start = "" if start is None else start
    period = "" if period is None else str(period)
    end = "" if end is None else end
    return base % (gage, period, start, end)


def getUSGSData(gage, end=None, start=None, period=None, urlFunc=usgsURL, si=False):
    """
    Retrieve USGS gage data for the given gage.
    :param gage: gage number
    :param end: end date (yyyy-mm-dd)
    :param start: start date
    :param period: number of days to retrieve data for
    :param urlFunc: url generator function (usually should be left as default) (arguments gage, start, end, period)
    :param si: if true, convert data from cfs/ft to cms/m
    :return: [(flow, stage)]
    """
    # identify flow and stage
    flown = "00060"
    stagen = "00065"
    url = urlFunc(gage=gage, start=start, end=end, period=period)
    with urlopen(url) as res:
        usgsBytes = res.read()
    usgs = usgsBytes.decode("utf-8")
    lines = usgs.split("\n")
    dataIx = 0
    while dataIx < len(lines):
        if lines[dataIx][0] != "#":  # Find first data line (not commented)
            break
        dataIx += 1
    rows = [l.split("\t") for l in lines[dataIx:]]
    flowcol = 0
    stagecol = 0
    for (ix, item) in enumerate(rows[0]):
        if item.endswith(flown):
            flowcol = ix
        if item.endswith(stagen):
            stagecol = ix
    volfactor = (12 / 39.37) ** 3 if si else 1
    stagefactor = 12 / 39.37 if si else 1
    return [
        # skip first 2 rows which are headers, not data, and make sure each row is long enough
        (float(row[flowcol]) * volfactor,
         float(row[stagecol])) * stagefactor for row in rows[2:] if len(row) > stagecol and
                                                                    len(row) > flowcol and
                                                                    len(row[flowcol]) > 0 and
                                                                    len(row[stagecol]) > 0
    ]


def prepareUSGSData(usgsData, flowcount=100, log=True):
    """
    Prepare flow and stage for use.  Returns a roughly evenly distributed set of flows across the relevant
    range.
    :param usgsData: usgs data as returned by getUSGSData - [(flow, stage)]
    :param flowcount: how many flows to return
    :param log: whether to evenly distribute logarithmically (alternative: linearly)
    :return: (flows, stages)
    """
    sortedData = sorted([u for u in usgsData if u[0] > 0], key=lambda d: d[0])  # sort by flow rate
    # Range: either largest / smallest or largest * smallest
    rng = sortedData[-1][0] / sortedData[0][0] if log else sortedData[-1][0] - sortedData[0][0]
    # flowcount - 1 steps
    step = rng ** (1 / (flowcount - 1)) if log else rng / (flowcount - 1)
    flow = [sortedData[0][0]]
    if flow[0] < 0.1:
        flow[0] = 0.1
    stage = [sortedData[0][1]]
    first = sortedData[0][0]

    # To avoid overlong numbers, since the HEC-RAS flow files are fixed-width, PyRASFile rounds to 1 decimal place
    first = first if first >= 0.1 else 0.1
    vals = [first * step ** ix if log else first + step + ix for ix in range(1, flowcount)]
    vals = [round(val, 1) for val in vals]
    vals = sorted(list(set(vals)))  # Only unique values, since rounding might introduce duplicates
    ixv = 0
    for (fl, st) in sortedData:
        if fl >= vals[ixv] and fl > flow[-1]:  # Make sure it's also greater than previous - no point in duplicates
            flow.append(fl)
            stage.append(st)
            ixv += 1
        if ixv >= len(vals):
            break
    return (flow, stage)


def singleStageFile(path):
    """
    Parse a single rating curve file with flow vs stage, assuming the flows are in the order of
    flow profiles in HEC-RAS which the user has entered and that the file is a CSV with column headers
    Flow and Stage.
    """
    lines = []
    with open(path) as f:
        lines = [line for line in f]
    lines = [[i.strip() for i in line.split(",")] for line in lines]
    stage = lines[0].index("Stage")
    flow = lines[0].index("Flow")
    return ([float(i[flow]) for i in lines[1:]], [float(i[stage]) for i in lines[1:]])



"""
Accept relevant inputs and put them into a useful format.

Copyright (C) 2020 Daniel Philippus
Full copyright notice located in main.py.
"""

from raspy_cal.default import Model
from raspy_cal.lowlevel import runSims
from raspy_cal.midlevel.eval import evaluate, minimized, fullEval, tests, evaluator
from raspy_cal.midlevel.params import paramSpec, genParams
from raspy_cal.frontend.display import evalTable, compareAllRatingCurves, nDisplay
from raspy_cal.midlevel.data import getUSGSData, prepareUSGSData, singleStageFile
from raspy_cal.midlevel.calibrators import nstageIteration, nstageSingleRun

from platypus import NSGAII, Problem, Real, nondominated # https://platypus.readthedocs.io/en/latest/getting-started.html#defining-constrained-problems
from urllib.request import urlopen

def parseConfigText(text, parsers):
    """
    Parse the text of a config file.  Format is Keyword: Value one per line.  Lines starting with # are
    ignored.
    :param text: the text of the config file.
    :param parsers: a dictionary of {keyword: function}, where function takes the string and returns
        the relevant value.  Lowercase keywords.
    :return: a dictionary of {keyword: value}
    """
    lines = [(line, line.index(":")) for line in text.split("\n") if not line.startswith("#")
             and not line == ""]
    # Before colon : after colon
    stringvals = { line[0][:line[1]].strip().lower() : line[0][line[1]+1:].strip() for line in lines }
    # So that values will be None if they aren't present in the file
    result = {k: None for k in parsers}
    for v in stringvals:
        if stringvals[v] != "":
            result[v] = parsers[v](stringvals[v])
    return result

def configSpecify(confPath, run = True):
    """
    Parse all of the arguments for specify and run it.
    :param confPath: path to the config file, or None to return example config file format
    :return: config values or example file format
    """
    id = lambda x: x
    toBool = lambda x: x in ["True", "true", "t", "T"]
    parsers = {
        "project": id,
        "stagef": id,
        "river": id,
        "reach": id,
        "rs": id,
        "nct": int,
        "outf": id,
        "plot": toBool,
        "auto": toBool,
        "evals": int,
        "metrics": lambda x: [i.strip() for i in x.split(",")],  # format: r2,rmse,pbias
        "filen": id,
        "slope": float,
        "usgs": id,
        "flowcount": int,
        "enddate": id,
        "startdate": id,
        "period": int,
        "si": toBool
    }
    if confPath is not None:
        with open(confPath) as f:
            data = f.read()
        vals = parseConfigText(data, parsers)
        if run:
            specify(
                project=vals["project"], stagef=vals["stagef"], river=vals["river"], reach=vals["reach"],
                rs=vals["rs"], nct=vals["nct"], outf=vals["outf"], plot=vals["plot"], auto=vals["auto"],
                evals=vals["evals"], metrics=vals["metrics"], fileN=vals["filen"], slope=vals["slope"],
                usgs=vals["usgs"], flowcount=vals["flowcount"], enddate=vals["enddate"], startdate=vals["startdate"],
                period=vals["period"], si=vals["si"]
            )
        return vals
    else:
        return """
# Example configuration file.  Fill out the values below to use and run with `python main.py <path>` or
# `raspy-cal.exe <path>` to use this config file.  Comment out lines you don't want to specify with `#`
# at the start of the line; information not provided will be requested when running as needed.  Not all
# information is required, e.g. if usgs is specified stagef will be ignored.
project: C:\\PathToHECRASProject\\project.prj
stagef: C:\\PathToStagefileIfUsed\\stagefile.csv
river: South Platte River
reach: Above Confluence Park
rs: 12345.0*
nct: 10
outf: C:\\PathToOutputFile\\outfile.csv
plot: True
auto: True
evals: 100
metrics: r2,rmse,pbias
filen: 04
slope: 0.025
usgs: 09429100
flowcount: 100
enddate: 2020-02-26
startdate: 2019-02-28
period: 500
si: False
"""

def specify(project = None, stagef = None, river = None, reach = None, rs = None, nct = None, outf = None,
            plot = None, auto = None, evals = None, metrics = None, fileN = None, slope = None, usgs = None,
            flowcount = None, enddate = None, startdate = None, period = None, correctDatum = None, si=None):
    """
    Select options and decide what to do.  All arguments are requested interactively if not specified.
    :param project: project path
    :param stagef: stage file path
    :param river: river name
    :param reach: reach name
    :param rs: river station for data collection
    :param nct: how many ns to test (per generation, if auto)
    :param outf: output file path or "" not to write one
    :param plot: to plot results
    :param auto: whether to use automatic optimization with NSGAII
    :param metrics: list of metrics to use
    :param fileN: flow file number to write to
    :param slope: slope for normal depth boundary condition
    :param usgs: USGS gage to use or "" not to use
    :param flowcount: how many flows to select if using USGS data
    :param enddate: end date for USGS
    :param startdate: start date for USGS
    :param period: period for USGS
    :param correctDatum: whether to adjust the datum
    :param si: use SI units
    """
    def getUSGS(usgs, flowcount, enddate, startdate, period, si):
        flowcount = int(input("Approx. how many flows to retrieve: ")) if flowcount is None else flowcount
        enddate = input("End date or leave blank for today: ") if enddate is None else enddate
        startdate = input("Start date or leave blank for 1 week ago or period: ") if startdate is None and\
                                                                                    (period is None or period == "")\
                                                                                    else startdate
        period = input("Period or leave blank for 1 week or start date: ") if period is None else period
        return prepareUSGSData(
            getUSGSData(usgs, enddate, startdate, period, si=si),
            flowcount
        )


    project = input("Enter project path (including .prj file): ") if project is None else project
    usgs = "" if stagef is not None else usgs
    usgs = input("USGS gage number or leave blank to use a stage file: ") if usgs is None else usgs
    if usgs == "":
        stagef = input("Enter path to stage file: ") if stagef is None else stagef
    (flow, stage) = singleStageFile(stagef) if usgs == "" else getUSGS(usgs, flowcount, enddate, startdate, period)
    outf = input("Enter output file path or nothing to not have one: ") if outf is None else outf
    fileN = input("Enter flow file number to write (default 01): ") if fileN is None else fileN
    fileN = "01" if fileN == "" else fileN
    slope = input("Enter slope for normal depth (default 0.001): ") if slope is None else slope
    slope = float(0.001) if slope == "" else float(slope)
    if metrics is None:
        metrics = [] if input("Enter Y to specify metrics: ") in ["Y", "y"] else None
        keys = list(tests.keys())
        if metrics == []:
            inp = ""
            while inp not in ["D", "d"]:
                print("Available metrics: %s" % keys)
                print("Selected metrics: %s" % metrics)
                inp = input("Enter a metric to add it or D if done: ")
                if inp in keys:
                    metrics.append(inp)
                elif inp not in ["D", "d"]:
                    print("Warning: entered metric is not an option.")
        if metrics == []:
            metrics = None
    river = input("River name: ") if river is None else river
    reach = input("Reach name: ") if reach is None else reach
    rs = input("River station: ") if rs is None else rs
    auto = input("Enter Y to use automatic calibration (default: interactive): ") in ["Y", "y"] if auto is None else auto
    plot = input("Enter N to not plot results (default: plot): ") not in ["N", "n"] if plot is None else plot
    nct = int(input("Number of n to test each iteration: ")) if nct is None else nct
    evals = int(input("How many evaluations to run? ")) if auto and evals is None else evals
    correctDatum = input("Enter Y to correct datum (default: no correction): ") in ["Y", "y"] if correctDatum is None else correctDatum
    si = input("Enter Y if HEC-RAS project and flow data are in SI units (default: US customary): ") in ["Y", "y"] if si is None else si
    model = Model(project)
    model.params.setSteadyFlows(river, reach, rs=None, flows=flow, slope=slope, fileN=fileN)

    if not auto:
        iterate(project = project, flow=flow, stage = stage, river = river, reach = reach, rs = rs, nct = nct,
                outf = outf, model = model, plot = plot, metrics = metrics, correctDatum = correctDatum, si=si)
    if auto:
        autoIterate(model = model, river = river, reach = reach, rs = rs, flow = flow, stage = stage, nct = nct,
                    plot = plot, outf = outf, metrics = metrics, evals = evals, correctDatum = correctDatum, si=si)

def iterate(project = None, flow = None, stage = None, river = None, reach = None, rs = None, nct = None,
            rand = None, outf = None, model = None, plot = None, metrics = None, correctDatum = None, si=False):
    """
    Iterate over n options until the user narrows it down to a good choice.  Note that providing an n of 0 will
    cause HEC-RAS to crash.
    :return: final ns
    """
    project = input("Enter project path (including .prj file): ") if project is None else project
    model = Model(project) if model is None else model
    river = input("River name: ") if river is None else river
    reach = input("Reach name: ") if reach is None else reach
    rs = input("River station: ") if rs is None else rs
    nct = int(input("Enter number of n to test each iteration: ")) if nct is None else nct
    rand = input("Enter Y to use random parameter generation: ") in ["y", "Y"] if rand is None else rand
    outf = input("Enter path to write output file to or nothing to not write it: ") if outf is None else outf
    plotpath = ".".join(outf.split(".")[:-1]) + ".png"
    plot = input("Plot results? Enter N to not plot: ") not in ["n", "N"] if plot is None else plot
    correctDatum = input("Enter Y to correct datum (default: no correction): ") in ["Y", "y"] if correctDatum is None else correctDatum
    cont = True
    while cont:
        nmin = float(input("Enter minimum n: "))
        nmax = float(input("Enter maximum n: "))
        best = nstageIteration(model, river, reach, rs, stage, nct, rand, nmin, nmax, metrics, correctDatum)
        # Show plot (if specified) but don't save anything
        nDisplay(best, flow, stage, None, None, plot, correctDatum, si)
        cont = input("Continue?  Q or q to quit and write results: ") not in ["q", "Q"]
        if not cont:
            # Save the plot and CSV
            nDisplay(best, flow, stage, plotpath, outf, False, correctDatum, si)

def autoIterate(model, river, reach, rs, flow, stage, nct, plot, outf, metrics, correctDatum, evals = None, si=False):
    """
    Automatically iterate with NSGA-II
    """
    keys = metrics  # ensure same order
    evalf = evaluator(stage, useTests = keys, correctDatum = correctDatum)
    evals = int(input("How many evaluations to run? ")) if evals is None else evals
    plotpath = ".".join(outf.split(".")[:-1]) + ".png"
    count = 1
    print("Running automatic calibration")
    def manningEval(vars):
        n = vars[0]
        metrics = minimized(
            nstageSingleRun(model, river, reach, rs, stage, n, keys, correctDatum)
        )
        values = [metrics[key] for key in keys]
        constraints = [-n, n - 1]
        nonlocal count
        print("Completed %d evaluations" % count)
        count += 1
        return values, constraints
    c_type = "<0"
    problem = Problem(1, len(keys), 2)  # 1 decision variable, len(keys) objectives, and 2 constraints
    problem.types[:] = Real(0.001, 1)  # range of decision variable
    problem.constraints[:] = c_type
    problem.function = manningEval

    algorithm = NSGAII(problem, population_size = nct)
    algorithm.run(evals)
    nondom = nondominated(algorithm.result) # nondom: list of Solutions - wanted value is variables[0]
    nondomNs = [sol.variables[0] for sol in nondom]
    results = runSims(model, nondomNs, river, reach, len(stage), range = [rs])
    resultPts = [(nondomNs[ix], [results[ix][rs][jx] for jx in range(1, len(stage) + 1)]) for ix in range(len(nondomNs))]
    metrics = [(res[0], evalf(res[1]), res[1]) for res in resultPts]
    nDisplay(metrics, flow, stage, plotpath, outf, plot, correctDatum, si)
    return metrics


def testrun():
    specify(
        project = "V:\\LosAngelesProjectsData\\HEC-RAS\\raspy\\DemoProject\\project.prj",
        stagef = "V:\\LosAngelesProjectsData\\HEC-RAS\\raspy_cal\\DemoStage.csv",
        river = "river1",
        reach = "reach1",
        rs = "200",
        # nct = 10,
        # rand = False,
        outf = "V:\\LosAngelesProjectsData\\HEC-RAS\\raspy_cal\\DemoOut.txt",
        plot = True
    )

if __name__ == "__main__":
    gn = "09423350"
    print(prepareUSGSData(getUSGSData(gn, period=365*2), log=False))










"""
Graphical front-end.  Just some simple inputs with which to run iterate and autoIterate.

Copyright (C) 2020 Daniel Philippus
Full copyright notice located in main.py.
"""

from raspy_cal.frontend.input import autoIterate, singleStageFile, configSpecify
from raspy_cal.midlevel.data import getUSGSData, prepareUSGSData
from raspy_cal.midlevel.calibrators import nstageIteration
from raspy_cal.frontend.display import evalTable, csv, nDisplay
from raspy_cal.midlevel.eval import tests
from raspy_cal.default import Model
import tkinter as tk
from tkinter import filedialog

# iteration(model, river, reach, rs, stage, flow, nct, rand, nmin, nmax, metrics, plot): [(n, metrics, sim)]
# autoIterate(model, river, reach, rs, flow, stage, nct, plot, outf, metrics, evals = None): [(n, metrics)]

defaultBase = "V:\\LosAngelesProjectsData\\HEC-RAS\\"

def browseButton(parent, f):
    return tk.Button(parent, text="Browse", command=lambda: f(filedialog.askopenfilename()))

class GUI(tk.Frame):
    def __init__(self, master = None):
        super().__init__(master)
        self.displayed = False
        self.master = master
        self.pack()
        self.createWidgets()

    def saveParameters(self):
        self.project = self.projectField.get()
        self.river = self.riverField.get()
        self.reach = self.reachField.get()
        self.rs = self.rsField.get()
        self.usgs = self.usgsField.get()
        self.stagef = self.stageField.get()
        (self.flow, self.stage) = singleStageFile(self.stagef) if self.usgs == "" else\
            prepareUSGSData(getUSGSData(self.usgs, period = 365 * 2))
        self.normalSlope = (lambda s: 0.001 if s == "" else float(s))(self.slopeField.get())
        self.fileN = (lambda n: "01" if n == "" else n)(self.fileNField.get())

        self.nct = int(self.nField.get())
        self.outf = self.outField.get()
        self.metrics = [key for key in self.keyChecks if self.keyChecks[key].get() == 1]
        self.plot = self.plotInt.get() == 1
        self.datum = self.datumInt.get() == 1
        self.si = self.siInt.get() == 1

        print("Parameters: %s" % [self.project, self.river, self.reach, self.rs, self.stagef,
                                  self.nct, self.outf, self.metrics, self.plot, self.datum])

    def mainMenu(self):
        self.iterFrame.destroy()
        self.entryFrame.pack(side="top")
        self.buttonFrame.pack(side="bottom")

    def selectRunType(self, runType):
        # runType: "auto" or "manual"
        self.saveParameters()
        self.model = Model(self.project)
        self.model.params.setSteadyFlows(self.river, self.reach, rs=None, flows=self.flow, slope=self.normalSlope,
                                         fileN=self.fileN)

        self.entryFrame.pack_forget()
        self.buttonFrame.pack_forget()
        self.iterFrame = tk.Frame(self)
        self.inputFrame = tk.Frame(self.iterFrame)
        if runType == "auto":
            self.autoInterface()
        elif runType == "manual":
            self.manualInterface()
        self.inputFrame.pack(side="top")
        self.iterFrame.pack(side="bottom")
        tk.Button(self.iterFrame, text="Return to Main Menu", command=self.mainMenu).pack(side="bottom")

    def autoInterface(self):
        self.evalsEntry = tk.Entry(self.inputFrame)
        tk.Label(self.inputFrame, text="How many evaluations to run?").grid(row=0)
        self.evalsEntry.grid(row=0, column=1)
        tk.Button(self.iterFrame, text="Run Automatic Calibration", command=self.runGenetic).pack(side="bottom")

    def manualInterface(self):
        self.nminEntry = tk.Entry(self.inputFrame)
        self.nmaxEntry = tk.Entry(self.inputFrame)
        self.randVar = tk.IntVar()
        self.randCheck = tk.Checkbutton(self.inputFrame, text="Random n distribution?", variable=self.randVar)

        fields = [
            ("Minimum n", self.nminEntry),
            ("Maximum n", self.nmaxEntry),
            ("Randomize n", self.randCheck)
        ]

        for (ix, (name, field)) in enumerate(fields):
            tk.Label(self.inputFrame, text=name).grid(row = ix)
            field.grid(row=ix, column=1)

        tk.Button(self.iterFrame, text="Run Simulations", command=self.runSims).pack(side="bottom")

    def runGenetic(self):
        print("Launching automatic calibration")
        self.evals = int(self.evalsEntry.get())
        self.result = autoIterate(self.model, self.river, self.reach, self.rs,
                                  self.flow, self.stage, self.nct, False, self.outf,
                                  self.metrics, self.datum, self.evals, self.si)
        self.displayResult()

    def runSims(self):
        self.nmin = float(self.nminEntry.get())
        self.nmax = float(self.nmaxEntry.get())
        self.rand = self.randVar.get() == 1
        self.result = nstageIteration(self.model, self.river, self.reach, self.rs, self.stage,
                                      self.nct, self.rand, self.nmin, self.nmax, self.metrics,
                                      self.datum)
        self.displayResult()

    def displayResult(self):
        if self.displayed:
            self.displayFrame.destroy()
        self.resultTable = evalTable([r[0] for r in self.result], [r[1] for r in self.result])
        self.displayFrame = tk.Frame(self.iterFrame)
        tk.Label(self.displayFrame, text=self.resultTable).pack(side="top")
        tk.Button(self.displayFrame, text="Save Results", command=self.writeResult).pack(side="bottom")
        self.displayFrame.pack(side="top")
        self.displayed = True
        if self.plot:
            nDisplay(self.result, self.flow, self.stage, plot=True, correctDatum=self.datum, si=self.si)

    def writeResult(self):
        self.plotpath = ".".join(self.outf.split(".")[:-1]) + ".png"
        nDisplay(self.result, self.flow, self.stage, csvpath=self.outf, plot=False, plotpath=self.plotpath,
                 correctDatum=self.datum, si=self.si)

    def loadConfig(self):
        vals = configSpecify(self.confField.get(), run=False)
        for k in self.pairs:
            if vals[k] is not None:
                self.pairs[k].delete(0, tk.END)
                self.pairs[k].insert(0, str(vals[k]))

    def saveConfig(self):
        out = [[k, self.pairs[k].get()] for k in self.pairs]
        file = self.confField.get()
        with open(file, "w") as f:
            f.write("\n".join([":".join(i) for i in out]))
    
    def setVal(self, entry):
        def setter(val):
            entry.delete(0, tk.END)
            entry.insert(0, val)
        return setter
            
    def browseField(self, parent):
        def setVal(entry, val):
            entry.delete(0, tk.END)
            entry.insert(0, val)
        browseFrame = tk.Frame(parent)
        entry = tk.Entry(browseFrame, width=100)
        browse = tk.Button(text="Browse", command=lambda: setVal(entry, tk.filedialog.askdirectory()))
        entry.grid(column=0, row=0)
        browse.grid(column=1, row=0)
        
        return (browseFrame, entry, browse)

    def createWidgets(self):
        self.keys = list(tests.keys())

        self.plotInt = tk.IntVar()
        self.datumInt = tk.IntVar()
        self.siInt = tk.IntVar()

        # Entries
        self.entryFrame = tk.Frame(self)
        self.confField = tk.Entry(self.entryFrame, width=100)
        self.saveConfigButton = tk.Button(self.entryFrame, text="Save Config", command=self.saveConfig)
        self.loadConfigButton = tk.Button(self.entryFrame, text="Load Config", command=self.loadConfig)
        self.projectField = tk.Entry(self.entryFrame, width=100)
        self.usgsField = tk.Entry(self.entryFrame, width = 100)
        self.riverField = tk.Entry(self.entryFrame, width=100)
        self.reachField = tk.Entry(self.entryFrame, width=100)
        self.rsField = tk.Entry(self.entryFrame, width=100)
        self.stageField = tk.Entry(self.entryFrame, width=100)
        self.slopeField = tk.Entry(self.entryFrame, width=100)
        self.fileNField = tk.Entry(self.entryFrame, width=100)
        self.nField = tk.Entry(self.entryFrame, width=100)
        self.outField = tk.Entry(self.entryFrame, width=100)
        self.metricField = tk.Frame(self.entryFrame)
        self.plotField = tk.Checkbutton(self.entryFrame, text="Plot?", variable=self.plotInt)
        self.datumField = tk.Checkbutton(self.entryFrame, text="Correct Datum?", variable=self.datumInt)
        self.siField = tk.Checkbutton(self.entryFrame, text="SI Units?", variable=self.siInt)

        self.pairs = {
            "project": self.projectField,
            "stagef": self.stageField,
            "river": self.riverField,
            "reach": self.reachField,
            "rs": self.rsField,
            "nct": self.nField,
            "outf": self.outField,
            "filen": self.fileNField,
            "slope": self.slopeField,
            "usgs": self.usgsField
        }

        self.keyChecks = {}

        for (ix, key) in enumerate(self.keys):
            var = tk.IntVar()
            chk = tk.Checkbutton(self.metricField, text=key, variable=var)
            chk.grid(row=0, column=ix)
            self.keyChecks[key] = var

        fields = [
            ("Config File Path (optional)", self.confField, browseButton(self.entryFrame, self.setVal(self.confField))),
            ("Save Config", self.saveConfigButton),
            ("Load Config", self.loadConfigButton),
            ("Project File Path", self.projectField, browseButton(self.entryFrame, self.setVal(self.projectField))),
            ("USGS Gage # (optional)", self.usgsField),
            ("River Name", self.riverField),
            ("Reach Name", self.reachField),
            ("River Station", self.rsField),
            ("Stage File Path", self.stageField),
            ("Slope for Normal Depth", self.slopeField),
            ("Flow file number to write (e.g. 01)", self.fileNField),
            ("# ns To Test", self.nField),
            ("Output File Path", self.outField, browseButton(self.entryFrame, self.setVal(self.outField))),
            ("Metrics", self.metricField),
            ("Plot", self.plotField),
            ("Datum", self.datumField),
            ("SI Units", self.siField)
        ]

        for (ix, vals) in enumerate(fields):
            name = vals[0]
            field = vals[1]
            tk.Label(self.entryFrame, text=name).grid(row=ix)
            field.grid(row=ix, column=1)
            if len(vals) == 3:
                vals[2].grid(row=ix, column=2)

        self.entryFrame.pack(side="top")

        self.buttonFrame = tk.Frame(self)
        self.runAuto = tk.Button(self.buttonFrame, text = "Automatic Calibration", command=lambda: self.selectRunType("auto"))
        self.runAuto.pack(side="left")
        self.runItv = tk.Button(self.buttonFrame, text = "Interactive Calibration", command = lambda: self.selectRunType("manual"))
        self.runItv.pack(side = "right")
        self.buttonFrame.pack(side = "bottom")


LICENSE = """Raspy-Cal Automatic Calibrator
Copyright (C) 2020 Daniel Philippus
This program comes with ABSOLUTELY NO WARRANTY.  This is free software, and you are
welcome to redistribute it under certain conditions.  For details, see the LICENSE
file at github.com/quantum-dan/raspy-cal.  This software is released under the GNU
General Public License v3.
"""

def main():
    root = tk.Tk()
    mainframe = tk.Frame(root)
    tk.Label(mainframe, text=LICENSE).pack(side="bottom")
    gui = GUI(mainframe)
    mainframe.master.title("Raspy-Cal Calibrator")
    mainframe.pack()
    gui.mainloop()

if __name__ == "__main__":
    main()





"""
Non-interactive components of calibration functions.

Copyright (C) 2020 Daniel Philippus.
Full copyright notice located in main.py.
"""

from raspy_cal.midlevel.params import paramSpec, genParams
from raspy_cal.midlevel.eval import evaluate, evaluator
from raspy_cal.lowlevel import runSims

def nstageIteration(model, river, reach, rs, stage, nct, rand, nmin, nmax, metrics, correctDatum):
    """
    Run one test.
    :param model: HEC-RAS model
    :param river: river name
    :param reach: reach name
    :param rs: river station
    :param stage: list of stages
    :param nct: number of ns to test
    :param rand: whether to use random ns
    :param nmin: minimum n
    :param nmax: maximum n
    :param metrics: list of metrics to use
    :return: [(n, metrics, sim)]
    """
    return multiRunner(model,
                  nstageMultiRunspec(river, reach, rs, len(stage)),
                  paramSpec("n", nmin, nmax, nct, rand),
                  nstageMultiEvaluator(stage, metrics, correctDatum))

def nstageSingleRun(model, river, reach, rs, stage, n, metrics, correctDatum):
    return singleRunner(model, nstageSingleRunspec(river, reach, rs, len(stage)),
                        {"n": n}, nstageSingleEvaluator(stage, metrics, correctDatum))

def nstageMultiRunspec(river, reach, rs, pcount):
    """
    Generates runspec function for roughness coefficient and stage.
    :param pcount: number of flow profiles
    :return: runspec function which returns [(n, simulated stage)]
    """
    def runspec(model, pspec):
        ns = [round(n, 3) for n in genParams([pspec], dicts=False)]
        results = runSims(model, ns, river, reach, pcount, range=[rs])
        return [(ns[ix], [results[ix][rs][jx] for jx in range(1, pcount + 1)]) for ix in range(len(ns))]
    return runspec

def nstageMultiEvaluator(stage, metrics, correctDatum):
    """
    Generates evaluator function for roughness coefficient and stage.
    :param stage: observed stage
    :param metrics: list of metrics to use
    :return: evaluator function which returns [(n, metrics, sim)]
    """
    def evtr(result):
        return evaluate(stage, result, correctDatum, metrics=metrics, n=len(result)//3)
    return evtr

def nstageSingleRunspec(river, reach, rs, pcount):
    """
    Generates runspec function for a single roughness coefficient and stage.
    :param pcount: number of flow profiles
    :return: runspec function which returns simulated stage
    """
    def runspec(model, pset):
        n = pset["n"]
        result = runSims(model, [n], river, reach, pcount, range=[rs])
        return [result[0][rs][ix] for ix in range(1, pcount+1)]
    return runspec

def nstageSingleEvaluator(stage, metrics, correctDatum):
    """
    Generates evaluator function for roughness coefficient and stage.
    :param stage: observed stage
    :param metrics: list of metrics to use
    :return: metrics dictionary
    """
    return evaluator(stage, correctDatum, metrics)

def multiRunner(model, runspec, pspec, evaluator):
    """
    Generic iteration function independent of the internal details of runspec, pspec, and evaluator.
    For use with interactive mode (see nstageRunspec).
    :param model: model as required by runspec
    :param runspec: function accepting arguments model and pspec to run tests as specified
    :param pspec: parameter specification required by runspec, returned by paramSpec
    :param evaluator: evaluator function that takes runspec results and returns a list of the best results:
        [(result, metrics dictionary, ...)] (further results [...] optional, e.g. stage timeseries)
    :return: results from evaluator
    """
    results = runspec(model, pspec)
    return evaluator(results)

def singleRunner(model, runspec, pset, evaluator):
    """
    Generic single-parameter-test function independent of the internal details of runspec, pspec, and evaluator.
    For use with automatic mode.
    :param model: model
    :param runspec: function accepting arguments model and pset to run a test as specified
    :param pset: parameter or set of parameters for the run (dictionary of values)
    :param evaluator: evaluator function that takes runspec results and returns a metrics dictionary
    :return: metrics from evaluator
    """
    result = runspec(model, pset)
    return evaluator(result)






"""
Low-level functionality that provides the infrastructure for optimization work.  This is a fairly
simple wrapper around the provided API.  See README.md for expected API functionality.

Copyright (C) 2020 Daniel Philippus
Full copyright notice located in main.py.
"""

STAGE = 0
VELOCITY = 1
ALL = -1

def runSims(model, mannings, river, reach, nprofs, range = None, retrieve = STAGE, log = True):
    """
    Run simulations and return the data.
    :param model: model API, already initialized appropriately
    :param mannings: list of Manning's n to test.  Each n can be a dictionary, list, or number -- see README.
    :param river: river to test
    :param reach: reach to test
    :param nprofs: number of flow profiles
    :param range: list of river stations to use, if specified.  Otherwise, the whole reach
    :param retrieve: STAGE, VELOCITY or ALL (0, 1, -1 respectively).  What data to retrieve.
    :return: list of the result data in order of the params used
    """
    out = []
    count = 1
    for n in mannings:
        if log:
            print("Running iteration")
        model.params.modifyN(n, river, reach)
        model.ops.compute(wait = True)
        # Below is repetitive, but it would introduce a lot of extra complexity to make it work as a function, I think
        if retrieve == STAGE:
            if range is None:
                out.append(model.data.stage(river, reach, None, nprofs))
            else:
                out.append({rs: model.data.stage(river, reach, rs, nprofs) for rs in range})
        elif retrieve == VELOCITY:
            if range is None:
                out.append(model.data.velocity(river, reach, None, nprofs))
            else:
                out.append({rs: model.data.velocity(river, reach, rs, nprofs) for rs in range})
        else:
            if range is None:
                out.append(model.data.allFlow(river, reach, None, nprofs))
            else:
                out.append({rs: model.data.allFlow(river, reach, rs, nprofs) for rs in range})
        if log:
            print("Completed %d simulations" % count)
            count += 1
    return out




"""
Doesn't actually do any processing, just provides a default model API from raspy.

model = Model(projectPath), then use that as model argument elsewhere - no direct interaction with raspy needed.

Copyright (C) 2020 Daniel Philippus
Full copyright notice located in main.py.
"""

from raspy_auto import Ras, API

def Model(projectPath):
    return API(Ras(projectPath))